"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Class that represents the catcli catalog
"""

import os
import pickle
from anytree.exporter import JsonExporter
from anytree.importer import JsonImporter

# local imports
import catcli.utils as utils
from catcli.logger import Logger


class Catalog:

    def __init__(self, path, pickle=False, debug=False, force=False):
        '''
        @path: catalog path
        @pickle: use pickle
        @debug: debug mode
        @force: force overwrite if exists
        '''
        self.path = path
        self.debug = debug
        self.force = force
        self.metanode = None
        self.pickle = pickle

    def set_metanode(self, metanode):
        '''remove the metanode until tree is re-written'''
        self.metanode = metanode
        self.metanode.parent = None

    def restore(self):
        '''restore the catalog'''
        if not self.path:
            return None
        if not os.path.exists(self.path):
            return None
        if self.pickle:
            return self._restore_pickle()
        return self._restore_json(open(self.path, 'r').read())

    def save(self, node):
        '''save the catalog'''
        if not self.path:
            Logger.err('Path not defined')
            return False
        d = os.path.dirname(self.path)
        if d and not os.path.exists(d):
            os.makedirs(d)
        elif os.path.exists(self.path) and not self.force:
            if not utils.ask('Update catalog \"{}\"'.format(self.path)):
                Logger.info('Catalog not saved')
                return False
        if d and not os.path.exists(d):
            Logger.err('Cannot write to \"{}\"'.format(d))
            return False
        if self.metanode:
            self.metanode.parent = node
        if self.pickle:
            return self._save_pickle(node)
        return self._save_json(node)

    def _debug(self, text):
        if not self.debug:
            return
        Logger.debug(text)

    def _save_pickle(self, node):
        '''pickle the catalog'''
        pickle.dump(node, open(self.path, 'wb'))
        self._debug('Catalog saved to pickle \"{}\"'.format(self.path))
        return True

    def _restore_pickle(self):
        '''restore the pickled tree'''
        root = pickle.load(open(self.path, 'rb'))
        m = 'Catalog imported from pickle \"{}\"'.format(self.path)
        self._debug(m)
        return root

    def _save_json(self, node):
        '''export the catalog in json'''
        exp = JsonExporter(indent=2, sort_keys=True)
        with open(self.path, 'w') as f:
            exp.write(node, f)
        self._debug('Catalog saved to json \"{}\"'.format(self.path))
        return True

    def _restore_json(self, string):
        '''restore the tree from json'''
        imp = JsonImporter()
        root = imp.import_(string)
        self._debug('Catalog imported from json \"{}\"'.format(self.path))
        return root



"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Logging helper
"""

import sys


class Logger:

    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    PURPLE = '\033[1;35m'
    BLUE = '\033[94m'
    GRAY = '\033[0;37m'
    MAGENTA = '\033[95m'
    RESET = '\033[0m'
    EMPH = '\033[33m'
    BOLD = '\033[1m'
    UND = '\033[4m'

    STORAGE = 'storage'
    ARCHIVE = 'archive'
    NBFILES = 'nbfiles'

    def no_color():
        Logger.RED = ''
        Logger.GREEN = ''
        Logger.YELLOW = ''
        Logger.PURPLE = ''
        Logger.BLUE = ''
        Logger.GRAY = ''
        Logger.MAGENTA = ''
        Logger.RESET = ''
        Logger.EMPH = ''
        Logger.BOLD = ''
        Logger.UND = ''

    def fix_badchars(line):
        return line.encode('utf-8', 'ignore').decode('utf-8')

    ######################################################################
    # node specific output
    ######################################################################
    def storage(pre, name, args, attr):
        '''print a storage node'''
        end = ''
        if attr:
            end = ' {}({}){}'.format(Logger.GRAY, attr, Logger.RESET)
        s = '{}{}{}{}:'.format(pre, Logger.UND, Logger.STORAGE, Logger.RESET)
        s += ' {}{}{}{}\n'.format(Logger.PURPLE,
                                  Logger.fix_badchars(name),
                                  Logger.RESET, end)
        s += '  {}{}{}'.format(Logger.GRAY, args, Logger.RESET)
        sys.stdout.write('{}\n'.format(s))

    def file(pre, name, attr):
        '''print a file node'''
        s = '{}{}'.format(pre, Logger.fix_badchars(name))
        s += ' {}[{}]{}'.format(Logger.GRAY, attr, Logger.RESET)
        sys.stdout.write('{}\n'.format(s))

    def dir(pre, name, depth='', attr=None):
        '''print a directory node'''
        end = []
        if depth != '':
            end.append('{}:{}'.format(Logger.NBFILES, depth))
        if attr:
            end.append(' '.join(['{}:{}'.format(x, y) for x, y in attr]))
        if end:
            end = ' [{}]'.format(', '.join(end))
        s = '{}{}{}{}'.format(pre, Logger.BLUE,
                              Logger.fix_badchars(name), Logger.RESET)
        s += '{}{}{}'.format(Logger.GRAY, end, Logger.RESET)
        sys.stdout.write('{}\n'.format(s))

    def arc(pre, name, archive):
        s = '{}{}{}{}'.format(pre, Logger.YELLOW,
                              Logger.fix_badchars(name), Logger.RESET)
        s += ' {}[{}:{}]{}'.format(Logger.GRAY, Logger.ARCHIVE,
                                   archive, Logger.RESET)
        sys.stdout.write('{}\n'.format(s))

    ######################################################################
    # generic output
    ######################################################################
    def out(string):
        '''to stdout no color'''
        string = Logger.fix_badchars(string)
        sys.stdout.write('{}\n'.format(string))

    def out_err(string):
        '''to stderr no color'''
        string = Logger.fix_badchars(string)
        sys.stderr.write('{}\n'.format(string))

    def debug(string):
        '''to stderr no color'''
        string = Logger.fix_badchars(string)
        sys.stderr.write('[DBG] {}\n'.format(string))

    def info(string):
        '''to stdout in color'''
        string = Logger.fix_badchars(string)
        s = '{}{}{}'.format(Logger.MAGENTA, string, Logger.RESET)
        sys.stdout.write('{}\n'.format(s))

    def err(string):
        '''to stderr in RED'''
        string = Logger.fix_badchars(string)
        s = '{}{}{}'.format(Logger.RED, string, Logger.RESET)
        sys.stderr.write('{}\n'.format(s))

    def progr(string):
        '''print progress'''
        string = Logger.fix_badchars(string)
        sys.stderr.write('{}\r'.format(string))
        sys.stderr.flush()

    def bold(string):
        '''make it bold'''
        string = Logger.fix_badchars(string)
        return '{}{}{}'.format(Logger.BOLD, string, Logger.RESET)

    def flog(path, string, append=True):
        string = Logger.fix_badchars(string)
        mode = 'w'
        if append:
            mode = 'a'
        with open(path, mode) as f:
            f.write(string)



"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Class that represents a node in the catalog tree
"""

import os
import anytree
import shutil
import time

# local imports
from . import __version__ as VERSION
import catcli.utils as utils
from catcli.logger import Logger
from catcli.decomp import Decomp

'''
There are 4 types of node:
    * "top" node representing the top node (generic node)
    * "storage" node representing a storage
    * "dir" node representing a directory
    * "file" node representing a file
'''


class Noder:

    TOPNAME = 'top'
    METANAME = 'meta'
    TYPE_TOP = 'top'
    TYPE_FILE = 'file'
    TYPE_DIR = 'dir'
    TYPE_ARC = 'arc'
    TYPE_STORAGE = 'storage'
    TYPE_META = 'meta'
    CSV_HEADER = ('name,type,path,size,indexed_at,'
                  'maccess,md5,nbfiles,free_space,'
                  'total_space,meta')

    def __init__(self, debug=False, sortsize=False, arc=False):
        '''
        @debug: debug mode
        @sortsize: sort nodes by size
        @arch: handle archive
        '''
        self.hash = True
        self.debug = debug
        self.sortsize = sortsize
        self.arc = arc
        if self.arc:
            self.decomp = Decomp()

    def get_storage_names(self, top):
        '''return a list of all storage names'''
        return [x.name for x in list(top.children)]

    def get_storage_node(self, top, name, path=None):
        '''
        return the storage node if any
        if path is submitted, it will update the media info
        '''
        found = None
        for n in top.children:
            if n.type != self.TYPE_STORAGE:
                continue
            if n.name == name:
                found = n
                break
        if found and path and os.path.exists(path):
            found.free = shutil.disk_usage(path).free
            found.total = shutil.disk_usage(path).total
            found.ts = int(time.time())
        return found

    def get_node(self, top, path, quiet=False):
        '''get the node by internal tree path'''
        r = anytree.resolver.Resolver('name')
        try:
            p = os.path.basename(path)
            return r.get(top, p)
        except anytree.resolver.ChildResolverError:
            if not quiet:
                Logger.err('No node at path \"{}\"'.format(p))
            return None

    def get_node_if_changed(self, top, path, treepath):
        '''
        return the node (if any) and if it has changed
        @top: top node (storage)
        @path: abs path to file
        @treepath: rel path from indexed directory
        '''
        treepath = treepath.lstrip(os.sep)
        node = self.get_node(top, treepath, quiet=True)
        # node does not exist
        if not node:
            self._debug('\tchange: node does not exist')
            return None, True
        if os.path.isdir(path):
            return node, False
        # force re-indexing if no maccess
        maccess = os.path.getmtime(path)
        if not self._has_attr(node, 'maccess') or \
                not node.maccess:
            self._debug('\tchange: no maccess found')
            return node, True
        # maccess changed
        old_maccess = node.maccess
        if float(maccess) != float(old_maccess):
            self._debug('\tchange: maccess changed for \"{}\"'.format(path))
            return node, True
        # test hash
        if self.hash and node.md5:
            md5 = self._get_hash(path)
            if md5 != node.md5:
                m = '\tchange: checksum changed for \"{}\"'.format(path)
                self._debug(m)
                return node, True
        self._debug('\tchange: no change for \"{}\"'.format(path))
        return node, False

    def _rec_size(self, node, store=True):
        '''
        recursively traverse tree and return size
        @store: store the size in the node
        '''
        if node.type == self.TYPE_FILE:
            self._debug('getting node size for \"{}\"'.format(node.name))
            return node.size
        m = 'getting node size recursively for \"{}\"'.format(node.name)
        self._debug(m)
        size = 0
        for i in node.children:
            if node.type == self.TYPE_DIR:
                sz = self._rec_size(i, store=store)
                if store:
                    i.size = sz
                size += sz
            if node.type == self.TYPE_STORAGE:
                sz = self._rec_size(i, store=store)
                if store:
                    i.size = sz
                size += sz
            else:
                continue
        if store:
            node.size = size
        return size

    def rec_size(self, node):
        '''recursively traverse tree and store dir size'''
        return self._rec_size(node, store=True)

    ###############################################################
    # public helpers
    ###############################################################
    def format_storage_attr(self, attr):
        '''format the storage attr for saving'''
        if not attr:
            return ''
        if type(attr) is list:
            return ', '.join(attr)
        attr = attr.rstrip()
        return attr

    def set_hashing(self, val):
        '''hash files when indexing'''
        self.hash = val

    ###############################################################
    # node creationg
    ###############################################################
    def new_top_node(self):
        '''create a new top node'''
        return anytree.AnyNode(name=self.TOPNAME, type=self.TYPE_TOP)

    def update_metanode(self, top):
        '''create or update meta node information'''
        meta = self._get_meta_node(top)
        epoch = int(time.time())
        if not meta:
            attr = {}
            attr['created'] = epoch
            attr['created_version'] = VERSION
            meta = anytree.AnyNode(name=self.METANAME, type=self.TYPE_META,
                                   attr=attr)
        meta.attr['access'] = epoch
        meta.attr['access_version'] = VERSION
        return meta

    def _get_meta_node(self, top):
        '''return the meta node if any'''
        try:
            return next(filter(lambda x: x.type == self.TYPE_META,
                        top.children))
        except StopIteration:
            return None

    def file_node(self, name, path, parent, storagepath):
        '''create a new node representing a file'''
        if not os.path.exists(path):
            Logger.err('File \"{}\" does not exist'.format(path))
            return None
        path = os.path.abspath(path)
        try:
            st = os.lstat(path)
        except OSError as e:
            Logger.err('OSError: {}'.format(e))
            return None
        md5 = None
        if self.hash:
            md5 = self._get_hash(path)
        relpath = os.sep.join([storagepath, name])

        maccess = os.path.getmtime(path)
        n = self._node(name, self.TYPE_FILE, relpath, parent,
                       size=st.st_size, md5=md5, maccess=maccess)
        if self.arc:
            ext = os.path.splitext(path)[1][1:]
            if ext.lower() in self.decomp.get_formats():
                self._debug('{} is an archive'.format(path))
                names = self.decomp.get_names(path)
                self.list_to_tree(n, names)
            else:
                self._debug('{} is NOT an archive'.format(path))
        return n

    def dir_node(self, name, path, parent, storagepath):
        '''create a new node representing a directory'''
        path = os.path.abspath(path)
        relpath = os.sep.join([storagepath, name])
        maccess = os.path.getmtime(path)
        return self._node(name, self.TYPE_DIR, relpath,
                          parent, maccess=maccess)

    def clean_not_flagged(self, top):
        '''remove any node not flagged and clean flags'''
        cnt = 0
        for node in anytree.PreOrderIter(top):
            if node.type != self.TYPE_FILE and node.type != self.TYPE_DIR:
                continue
            if self._clean(node):
                cnt += 1
        return cnt

    def flag(self, node):
        '''flag a node'''
        node.flag = True

    def _clean(self, node):
        '''remove node if not flagged'''
        if not self._has_attr(node, 'flag') or \
                not node.flag:
            node.parent = None
            return True
        del node.flag
        return False

    def storage_node(self, name, path, parent, attr=None):
        '''create a new node representing a storage'''
        path = os.path.abspath(path)
        free = shutil.disk_usage(path).free
        total = shutil.disk_usage(path).total
        epoch = int(time.time())
        return anytree.AnyNode(name=name, type=self.TYPE_STORAGE, free=free,
                               total=total, parent=parent, attr=attr, ts=epoch)

    def archive_node(self, name, path, parent, archive):
        '''crete a new node for archive data'''
        return anytree.AnyNode(name=name, type=self.TYPE_ARC, relpath=path,
                               parent=parent, size=0, md5=None,
                               archive=archive)

    def _node(self, name, type, relpath, parent,
              size=None, md5=None, maccess=None):
        '''generic node creation'''
        return anytree.AnyNode(name=name, type=type, relpath=relpath,
                               parent=parent, size=size,
                               md5=md5, maccess=maccess)

    ###############################################################
    # printing
    ###############################################################
    def _node_to_csv(self, node, sep=',', raw=False):
        '''
        print a node to csv
        @node: the node to consider
        @sep: CSV separator character
        @raw: print raw size rather than human readable
        '''
        if not node:
            return ''
        if node.type == self.TYPE_TOP:
            return ''

        out = []
        if node.type == self.TYPE_STORAGE:
            # handle storage
            out.append(node.name)   # name
            out.append(node.type)   # type
            out.append('')          # fake full path
            sz = self._rec_size(node, store=False)
            out.append(utils.size_to_str(sz, raw=raw))  # size
            out.append(utils.epoch_to_str(node.ts))  # indexed_at
            out.append('')  # fake maccess
            out.append('')  # fake md5
            out.append(str(len(node.children)))  # nbfiles
            # fake free_space
            out.append(utils.size_to_str(node.free, raw=raw))
            # fake total_space
            out.append(utils.size_to_str(node.total, raw=raw))
            out.append(node.attr)  # meta
        else:
            # handle other nodes
            out.append(node.name.replace('"', '""'))  # name
            out.append(node.type)  # type
            parents = self._get_parents(node)
            storage = self._get_storage(node)
            fullpath = os.path.join(storage.name, parents)
            out.append(fullpath.replace('"', '""'))  # full path

            out.append(utils.size_to_str(node.size, raw=raw))  # size
            out.append(utils.epoch_to_str(storage.ts))  # indexed_at
            if self._has_attr(node, 'maccess'):
                out.append(utils.epoch_to_str(node.maccess))  # maccess
            else:
                out.append('')  # fake maccess
            if node.md5:
                out.append(node.md5)  # md5
            else:
                out.append('')  # fake md5
            if node.type == self.TYPE_DIR:
                out.append(str(len(node.children)))  # nbfiles
            else:
                out.append('')  # fake nbfiles
            out.append('')  # fake free_space
            out.append('')  # fake total_space
            out.append('')  # fake meta

        line = sep.join(['"' + o + '"' for o in out])
        if len(line) > 0:
            Logger.out(line)

    def _print_node(self, node, pre='', withpath=False,
                    withdepth=False, withstorage=False,
                    recalcparent=False, raw=False):
        '''
        print a node
        @node: the node to print
        @pre: string to print before node
        @withpath: print the node path
        @withdepth: print the node depth info
        @withstorage: print the node storage it belongs to
        @recalcparent: get relpath from tree instead of relpath field
        @raw: print raw size rather than human readable
        '''
        if node.type == self.TYPE_TOP:
            # top node
            Logger.out('{}{}'.format(pre, node.name))
        elif node.type == self.TYPE_FILE:
            # node of type file
            name = node.name
            if withpath:
                if recalcparent:
                    name = os.sep.join([self._get_parents(node.parent), name])
                else:
                    name = node.relpath
            name = name.lstrip(os.sep)
            if withstorage:
                storage = self._get_storage(node)
            attr = ''
            if node.md5:
                attr = ', md5:{}'.format(node.md5)
            sz = utils.size_to_str(node.size, raw=raw)
            compl = 'size:{}{}'.format(sz, attr)
            if withstorage:
                compl += ', storage:{}'.format(Logger.bold(storage.name))
            Logger.file(pre, name, compl)
        elif node.type == self.TYPE_DIR:
            # node of type directory
            name = node.name
            if withpath:
                if recalcparent:
                    name = os.sep.join([self._get_parents(node.parent), name])
                else:
                    name = node.relpath
            name = name.lstrip(os.sep)
            depth = ''
            if withdepth:
                depth = len(node.children)
            if withstorage:
                storage = self._get_storage(node)
            attr = []
            if node.size:
                attr.append(['totsize', utils.size_to_str(node.size, raw=raw)])
            if withstorage:
                attr.append(['storage', Logger.bold(storage.name)])
            Logger.dir(pre, name, depth=depth, attr=attr)
        elif node.type == self.TYPE_STORAGE:
            # node of type storage
            hf = utils.size_to_str(node.free, raw=raw)
            ht = utils.size_to_str(node.total, raw=raw)
            nbchildren = len(node.children)
            freepercent = '{:.1f}%'.format(
                node.free * 100 / node.total
            )
            # get the date
            dt = ''
            if self._has_attr(node, 'ts'):
                dt = 'date:'
                dt += '{}'.format(utils.epoch_to_str(node.ts))
            ds = ''
            # the children size
            sz = self._rec_size(node, store=False)
            sz = utils.size_to_str(sz, raw=raw)
            ds = 'totsize:' + '{}'.format(sz)
            # format the output
            name = '{}'.format(node.name)
            args = [
                'nbfiles:' + '{}'.format(nbchildren),
                ds,
                'free:{}'.format(freepercent),
                'du:' + '{}/{}'.format(hf, ht),
                dt]
            Logger.storage(pre,
                           name,
                           '{}'.format(' | '.join(args)),
                           node.attr)
        elif node.type == self.TYPE_ARC:
            # archive node
            if self.arc:
                Logger.arc(pre, node.name, node.archive)
        else:
            Logger.err('bad node encountered: {}'.format(node))

    def print_tree(self, node, style=anytree.ContRoundStyle(),
                   fmt='native', header=False, raw=False):
        '''
        print the tree similar to unix tool "tree"
        @node: start node
        @style: when fmt=native, defines the tree style
        @fmt: output format
        @header: when fmt=csv, print the header
        @raw: print the raw size rather than human readable
        '''
        if fmt == 'native':
            rend = anytree.RenderTree(node, childiter=self._sort_tree)
            for pre, fill, node in rend:
                self._print_node(node, pre=pre, withdepth=True, raw=raw)
        elif fmt == 'csv':
            self._to_csv(node, with_header=header, raw=raw)

    def _to_csv(self, node, with_header=False, raw=False):
        '''print the tree to csv'''
        rend = anytree.RenderTree(node, childiter=self._sort_tree)
        if with_header:
            Logger.out(self.CSV_HEADER)
        for _, _, node in rend:
            self._node_to_csv(node, raw=raw)

    def to_dot(self, node, path='tree.dot'):
        '''export to dot for graphing'''
        anytree.exporter.DotExporter(node).to_dotfile(path)
        Logger.info('dot file created under \"{}\"'.format(path))
        return 'dot {} -T png -o /tmp/tree.png'.format(path)

    ###############################################################
    # searching
    ###############################################################
    def find_name(self, root, key,
                  script=False, directory=False,
                  startpath=None, parentfromtree=False,
                  fmt='native', raw=False):
        '''
        find files based on their names
        @script: output script
        @directory: only search for directories
        @startpath: node to start with
        @parentfromtree: get path from parent instead of stored relpath
        @fmt: output format
        '''
        self._debug('searching for \"{}\"'.format(key))
        start = root
        if startpath:
            start = self.get_node(root, startpath)
        self.term = key
        found = anytree.findall(start, filter_=self._find_name)
        paths = []
        for f in found:
            if f.type == self.TYPE_STORAGE:
                # ignore storage nodes
                continue
            if directory and f.type != self.TYPE_DIR:
                # ignore non directory
                continue

            # print the node
            if fmt == 'native':
                self._print_node(f, withpath=True,
                                 withdepth=True,
                                 withstorage=True,
                                 recalcparent=parentfromtree,
                                 raw=raw)
            elif fmt == 'csv':
                self._node_to_csv(f, raw=raw)

            if parentfromtree:
                paths.append(self._get_parents(f))
            else:
                paths.append(f.relpath)

        if script:
            tmp = ['${source}/' + x for x in paths]
            cmd = 'op=file; source=/media/mnt; $op {}'.format(' '.join(tmp))
            Logger.info(cmd)

        return found

    def _find_name(self, node):
        '''callback for finding files'''
        if self.term.lower() in node.name.lower():
            return True
        return False

    ###############################################################
    # climbing
    ###############################################################
    def walk(self, root, path, rec=False, fmt='native', raw=False):
        '''
        walk the tree for ls based on names
        @root: start node
        @rec: recursive walk
        @fmt: output format
        '''
        self._debug('walking path: \"{}\"'.format(path))

        r = anytree.resolver.Resolver('name')
        found = []
        try:
            found = r.glob(root, path)
            if len(found) < 1:
                # nothing found
                return []

            if rec:
                # print the entire tree
                self.print_tree(found[0].parent, fmt=fmt, raw=raw)
                return found

            # sort found nodes
            found = sorted(found, key=self._sort, reverse=self.sortsize)

            # print the parent
            if fmt == 'native':
                self._print_node(found[0].parent,
                                 withpath=False, withdepth=True, raw=raw)
            elif fmt == 'csv':
                self._node_to_csv(found[0].parent, raw=raw)

            # print all found nodes
            for f in found:
                if fmt == 'native':
                    self._print_node(f, withpath=False,
                                     pre='- ',
                                     withdepth=True,
                                     raw=raw)
                elif fmt == 'csv':
                    self._node_to_csv(f, raw=raw)

        except anytree.resolver.ChildResolverError:
            pass
        return found

    ###############################################################
    # tree creation
    ###############################################################
    def _add_entry(self, name, top, resolv):
        '''add an entry to the tree'''
        entries = name.rstrip(os.sep).split(os.sep)
        if len(entries) == 1:
            self.archive_node(name, name, top, top.name)
            return
        sub = os.sep.join(entries[:-1])
        f = entries[-1]
        try:
            parent = resolv.get(top, sub)
            parent = self.archive_node(f, name, parent, top.name)
        except anytree.resolver.ChildResolverError:
            self.archive_node(f, name, top, top.name)

    def list_to_tree(self, parent, names):
        '''convert list of files to a tree'''
        if not names:
            return
        r = anytree.resolver.Resolver('name')
        for name in names:
            name = name.rstrip(os.sep)
            self._add_entry(name, parent, r)

    ###############################################################
    # diverse
    ###############################################################
    def _sort_tree(self, items):
        '''sorting a list of items'''
        return sorted(items, key=self._sort, reverse=self.sortsize)

    def _sort(self, x):
        '''sort a list'''
        if self.sortsize:
            return self._sort_size(x)
        return self._sort_fs(x)

    def _sort_fs(self, n):
        '''sorting nodes dir first and alpha'''
        return (n.type, n.name.lstrip('\.').lower())

    def _sort_size(self, n):
        '''sorting nodes by size'''
        try:
            if not n.size:
                return 0
            return n.size
        except AttributeError:
            return 0

    def _get_storage(self, node):
        '''recursively traverse up to find storage'''
        if node.type == self.TYPE_STORAGE:
            return node
        return node.ancestors[1]

    def _has_attr(self, node, attr):
        return attr in node.__dict__.keys()

    def _get_parents(self, node):
        '''get all parents recursively'''
        if node.type == self.TYPE_STORAGE:
            return ''
        parent = self._get_parents(node.parent)
        if parent:
            return os.sep.join([parent, node.name])
        return node.name

    def _get_hash(self, path):
        """return md5 hash of node"""
        return utils.md5sum(path)

    def _debug(self, string):
        '''print debug'''
        if not self.debug:
            return
        Logger.debug(string)



"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

helpers
"""

import os
import hashlib
import tempfile
import subprocess
import datetime

# local imports
from catcli.logger import Logger


def md5sum(path):
    '''calculate md5 sum of a file'''
    p = os.path.realpath(path)
    if not os.path.exists(p):
        Logger.err('\nmd5sum - file does not exist: {}'.format(p))
        return None
    try:
        with open(p, mode='rb') as f:
            d = hashlib.md5()
            while True:
                buf = f.read(4096)
                if not buf:
                    break
                d.update(buf)
            return d.hexdigest()
    except PermissionError:
        pass
    except OSError as e:
        Logger.err('md5sum error: {}'.format(e))
    return None


def size_to_str(size, raw=True):
    '''convert size to string, optionally human readable'''
    div = 1024.
    suf = ['B', 'K', 'M', 'G', 'T', 'P']
    if raw or size < div:
        return '{}'.format(size)
    for i in suf:
        if size < div:
            return '{:.1f}{}'.format(size, i)
        size = size / div
    return '{:.1f}{}'.format(size, suf[-1])


def epoch_to_str(epoch):
    '''convert epoch to string'''
    if not epoch:
        return ''
    fmt = '%Y-%m-%d %H:%M:%S'
    t = datetime.datetime.fromtimestamp(float(epoch))
    return t.strftime(fmt)


def ask(question):
    '''ask the user what to do'''
    resp = input('{} [y|N] ? '.format(question))
    return resp.lower() == 'y'


def edit(string):
    '''edit the information with the default EDITOR'''
    string = string.encode('utf-8')
    EDITOR = os.environ.get('EDITOR', 'vim')
    with tempfile.NamedTemporaryFile(prefix='catcli', suffix='.tmp') as f:
        f.write(string)
        f.flush()
        subprocess.call([EDITOR, f.name])
        f.seek(0)
        new = f.read()
    return new.decode('utf-8')



"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Catcli generic compressed data lister
"""

import os
import tarfile
import zipfile


class Decomp:

    def __init__(self):
        self.ext = {
            'tar': self._tar,
            'tgz': self._tar,
            'gz': self._tar,
            'tar.gz': self._tar,
            'xz': self._tar,
            'tar.xz': self._tar,
            'lzma': self._tar,
            'tar.lzma': self._tar,
            'tlz': self._tar,
            'bz2': self._tar,
            'tar.bz2': self._tar,
            'zip': self._zip}

    def get_formats(self):
        '''return list of supported extensions'''
        return list(self.ext.keys())

    def get_names(self, path):
        '''get tree of compressed archive'''
        ext = os.path.splitext(path)[1][1:].lower()
        if ext in list(self.ext.keys()):
            return self.ext[ext](path)
        return None

    def _tar(self, path):
        '''return list of file names in tar'''
        if not tarfile.is_tarfile(path):
            return None
        tar = tarfile.open(path, "r")
        return tar.getnames()

    def _zip(self, path):
        '''return list of file names in zip'''
        if not zipfile.is_zipfile(path):
            return None
        z = zipfile.ZipFile(path)
        return z.namelist()



"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Catcli filesystem indexer
"""

import os

# local imports
from catcli.logger import Logger


class Walker:

    MAXLINE = 80 - 15

    def __init__(self, noder, hash=True, debug=False,
                 logpath=None):
        '''
        @noder: the noder to use
        @hash: calculate hash of nodes
        @debug: debug mode
        @logpath: path where to log catalog changes on reindex
        '''
        self.noder = noder
        self.hash = hash
        self.noder.set_hashing(self.hash)
        self.debug = debug
        self.lpath = logpath

    def index(self, path, parent, name, storagepath=''):
        '''
        index a directory and store in tree
        @path: path to index
        @parent: parent node
        @name: this stoarge name
        '''
        self._debug('indexing starting at {}'.format(path))
        if not parent:
            parent = self.noder.dir_node(name, path, parent)

        if os.path.islink(path):
            rel = os.readlink(path)
            ab = os.path.join(path, rel)
            if os.path.isdir(ab):
                return parent, 0

        cnt = 0
        for (root, dirs, files) in os.walk(path):
            for f in files:
                self._debug('found file {} under {}'.format(f, path))
                sub = os.path.join(root, f)
                if not os.path.exists(sub):
                    continue
                self._progress(f)
                self._debug('index file {}'.format(sub))
                n = self.noder.file_node(os.path.basename(f), sub,
                                         parent, storagepath)
                if n:
                    cnt += 1
            for d in dirs:
                self._debug('found dir {} under {}'.format(d, path))
                base = os.path.basename(d)
                sub = os.path.join(root, d)
                self._debug('index directory {}'.format(sub))
                if not os.path.exists(sub):
                    continue
                dummy = self.noder.dir_node(base, sub, parent, storagepath)
                if not dummy:
                    continue
                cnt += 1
                nstoragepath = os.sep.join([storagepath, base])
                if not storagepath:
                    nstoragepath = base
                _, cnt2 = self.index(sub, dummy, base, nstoragepath)
                cnt += cnt2
            break
        self._progress(None)
        return parent, cnt

    def reindex(self, path, parent, top):
        '''reindex a directory and store in tree'''
        cnt = self._reindex(path, parent, top)
        cnt += self.noder.clean_not_flagged(parent)
        return cnt

    def _reindex(self, path, parent, top, storagepath=''):
        '''
        reindex a directory and store in tree
        @path: directory path to re-index
        @top: top node (storage)
        @storagepath: rel path relative to indexed directory
        '''
        self._debug('reindexing starting at {}'.format(path))
        cnt = 0
        for (root, dirs, files) in os.walk(path):
            for f in files:
                self._debug('found file \"{}\" under {}'.format(f, path))
                sub = os.path.join(root, f)
                treepath = os.path.join(storagepath, f)
                reindex, n = self._need_reindex(parent, sub, treepath)
                if not reindex:
                    self._debug('\tskip file {}'.format(sub))
                    self.noder.flag(n)
                    continue
                self._log2file('update catalog for \"{}\"'.format(sub))
                n = self.noder.file_node(os.path.basename(f), sub,
                                         parent, storagepath)
                self.noder.flag(n)
                cnt += 1
            for d in dirs:
                self._debug('found dir \"{}\" under {}'.format(d, path))
                base = os.path.basename(d)
                sub = os.path.join(root, d)
                treepath = os.path.join(storagepath, d)
                reindex, dummy = self._need_reindex(parent, sub, treepath)
                if reindex:
                    self._log2file('update catalog for \"{}\"'.format(sub))
                    dummy = self.noder.dir_node(base, sub, parent, storagepath)
                    cnt += 1
                self.noder.flag(dummy)
                self._debug('reindexing deeper under {}'.format(sub))
                nstoragepath = os.sep.join([storagepath, base])
                if not storagepath:
                    nstoragepath = base
                cnt2 = self._reindex(sub, dummy, top, nstoragepath)
                cnt += cnt2
            break
        return cnt

    def _need_reindex(self, top, path, treepath):
        '''
        test if node needs re-indexing
        @top: top node (storage)
        @path: abs path to file
        @treepath: rel path from indexed directory
        '''
        cnode, changed = self.noder.get_node_if_changed(top, path, treepath)
        if not cnode:
            self._debug('\t{} does not exist'.format(path))
            return True, cnode
        if cnode and not changed:
            # ignore this node
            self._debug('\t{} has not changed'.format(path))
            return False, cnode
        if cnode and changed:
            # remove this node and re-add
            self._debug('\t{} has changed'.format(path))
            self._debug('\tremoving node {} for {}'.format(cnode.name, path))
            cnode.parent = None
        return True, cnode

    def _debug(self, string):
        '''print to debug'''
        if not self.debug:
            return
        Logger.debug(string)

    def _progress(self, string):
        '''print progress'''
        if self.debug:
            return
        if not string:
            # clean
            Logger.progr('{:80}'.format(' '))
            return
        if len(string) > self.MAXLINE:
            string = string[:self.MAXLINE] + '...'
        Logger.progr('indexing: {:80}'.format(string))

    def _log2file(self, string):
        '''log to file'''
        if not self.lpath:
            return
        line = '{}\n'.format(string)
        Logger.flog(self.lpath, line, append=True)



#!/usr/bin/env python3
# author: deadc0de6

"""
author: deadc0de6 (https://github.com/deadc0de6)
Copyright (c) 2017, deadc0de6

Catcli command line interface
"""

import sys
import os
import datetime
from docopt import docopt

# local imports
from . import __version__ as VERSION
from .logger import Logger
from .catalog import Catalog
from .walker import Walker
from .noder import Noder
from .utils import ask, edit

NAME = 'catcli'
CUR = os.path.dirname(os.path.abspath(__file__))
CATALOGPATH = '{}.catalog'.format(NAME)
GRAPHPATH = '/tmp/{}.dot'.format(NAME)
SEPARATOR = '/'
WILD = '*'

BANNER = """ +-+-+-+-+-+-+
 |c|a|t|c|l|i|
 +-+-+-+-+-+-+ v{}""".format(VERSION)

USAGE = """
{0}

Usage:
    {1} ls     [--catalog=<path>] [--format=<fmt>] [-aBCrVSs] [<path>]
    {1} find   [--catalog=<path>] [--format=<fmt>] [-aBCbdVsP] [--path=<path>] <term>
    {1} tree   [--catalog=<path>] [--format=<fmt>] [-aBCVSsH] [<path>]
    {1} index  [--catalog=<path>] [--meta=<meta>...] [-aBCcfnV] <name> <path>
    {1} update [--catalog=<path>] [-aBCcfnV] [--lpath=<path>] <name> <path>
    {1} rm     [--catalog=<path>] [-BCfV] <storage>
    {1} rename [--catalog=<path>] [-BCfV] <storage> <name>
    {1} edit   [--catalog=<path>] [-BCfV] <storage>
    {1} graph  [--catalog=<path>] [-BCV] [<path>]
    {1} print_supported_formats
    {1} help
    {1} --help
    {1} --version

Options:
    --catalog=<path>    Path to the catalog [default: {2}].
    --meta=<meta>       Additional attribute to store [default: ].
    -a --archive        Handle archive file [default: False].
    -B --no-banner      Do not display the banner [default: False].
    -b --script         Output script to manage found file(s) [default: False].
    -C --no-color       Do not output colors [default: False].
    -c --hash           Calculate md5 hash [default: False].
    -d --directory      Only directory [default: False].
    -F --format=<fmt>   Print format, see command \"print_supported_formats\" [default: native].
    -f --force          Do not ask when updating the catalog [default: False].
    -H --header         Print header on CSV format [default: False].
    -l --lpath=<path>   Path where changes are logged [default: ]
    -n --no-subsize     Do not store size of directories [default: False].
    -P --parent         Ignore stored relpath [default: True].
    -p --path=<path>    Start path.
    -r --recursive      Recursive [default: False].
    -s --raw-size       Print raw size rather than human readable [default: False].
    -S --sortsize       Sort by size, largest first [default: False].
    -V --verbose        Be verbose [default: False].
    -v --version        Show version.
    -h --help           Show this screen.
""".format(BANNER, NAME, CATALOGPATH)  # nopep8


def cmd_index(args, noder, catalog, top):
    path = args['<path>']
    name = args['<name>']
    hash = args['--hash']
    debug = args['--verbose']
    subsize = not args['--no-subsize']
    if not os.path.exists(path):
        Logger.err('\"{}\" does not exist'.format(path))
        return
    if name in noder.get_storage_names(top):
        try:
            if not ask('Overwrite storage \"{}\"'.format(name)):
                Logger.err('storage named \"{}\" already exist'.format(name))
                return
        except KeyboardInterrupt:
            Logger.err('aborted')
            return
        node = noder.get_storage_node(top, name)
        node.parent = None

    start = datetime.datetime.now()
    walker = Walker(noder, hash=hash, debug=debug)
    attr = noder.format_storage_attr(args['--meta'])
    root = noder.storage_node(name, path, parent=top, attr=attr)
    _, cnt = walker.index(path, root, name)
    if subsize:
        noder.rec_size(root)
    stop = datetime.datetime.now()
    Logger.info('Indexed {} file(s) in {}'.format(cnt, stop - start))
    if cnt > 0:
        catalog.save(top)


def cmd_update(args, noder, catalog, top):
    path = args['<path>']
    name = args['<name>']
    hash = args['--hash']
    logpath = args['--lpath']
    debug = args['--verbose']
    subsize = not args['--no-subsize']
    if not os.path.exists(path):
        Logger.err('\"{}\" does not exist'.format(path))
        return
    root = noder.get_storage_node(top, name, path=path)
    if not root:
        Logger.err('storage named \"{}\" does not exist'.format(name))
        return
    start = datetime.datetime.now()
    walker = Walker(noder, hash=hash, debug=debug,
                    logpath=logpath)
    cnt = walker.reindex(path, root, top)
    if subsize:
        noder.rec_size(root)
    stop = datetime.datetime.now()
    Logger.info('updated {} file(s) in {}'.format(cnt, stop - start))
    if cnt > 0:
        catalog.save(top)


def cmd_ls(args, noder, top):
    path = args['<path>']
    if not path:
        path = SEPARATOR
    if not path.startswith(SEPARATOR):
        path = SEPARATOR + path
    pre = '{}{}'.format(SEPARATOR, noder.TOPNAME)
    if not path.startswith(pre):
        path = pre + path
    if not path.endswith(SEPARATOR):
        path += SEPARATOR
    if not path.endswith(WILD):
        path += WILD
    found = noder.walk(top, path,
                       rec=args['--recursive'],
                       fmt=args['--format'],
                       raw=args['--raw-size'])
    if not found:
        Logger.err('\"{}\": nothing found'.format(args['<path>']))
    return found


def cmd_rm(args, noder, catalog, top):
    name = args['<storage>']
    node = noder.get_storage_node(top, name)
    if node:
        node.parent = None
        if catalog.save(top):
            Logger.info('Storage \"{}\" removed'.format(name))
    else:
        Logger.err('Storage named \"{}\" does not exist'.format(name))
    return top


def cmd_find(args, noder, top):
    fromtree = args['--parent']
    directory = args['--directory']
    startpath = args['--path']
    fmt = args['--format']
    raw = args['--raw-size']
    return noder.find_name(top, args['<term>'], script=args['--script'],
                           startpath=startpath, directory=directory,
                           parentfromtree=fromtree, fmt=fmt, raw=raw)


def cmd_tree(args, noder, top):
    path = args['<path>']
    fmt = args['--format']
    hdr = args['--header']
    raw = args['--raw-size']

    # find node to start with
    node = top
    if path:
        node = noder.get_node(top, path)

    if node:
        # print the tree
        noder.print_tree(node, fmt=fmt, header=hdr, raw=raw)


def cmd_graph(args, noder, top):
    path = args['<path>']
    if not path:
        path = GRAPHPATH
    cmd = noder.to_dot(top, path)
    Logger.info('create graph with \"{}\" (you need graphviz)'.format(cmd))


def cmd_rename(args, noder, catalog, top):
    storage = args['<storage>']
    new = args['<name>']
    storages = list(x.name for x in top.children)
    if storage in storages:
        node = next(filter(lambda x: x.name == storage, top.children))
        node.name = new
        if catalog.save(top):
            m = 'Storage \"{}\" renamed to \"{}\"'.format(storage, new)
            Logger.info(m)
    else:
        Logger.err('Storage named \"{}\" does not exist'.format(storage))
    return top


def cmd_edit(args, noder, catalog, top):
    storage = args['<storage>']
    storages = list(x.name for x in top.children)
    if storage in storages:
        node = next(filter(lambda x: x.name == storage, top.children))
        attr = node.attr
        if not attr:
            attr = ''
        new = edit(attr)
        node.attr = noder.format_storage_attr(new)
        if catalog.save(top):
            Logger.info('Storage \"{}\" edited'.format(storage))
    else:
        Logger.err('Storage named \"{}\" does not exist'.format(storage))
    return top


def banner():
    Logger.out_err(BANNER)
    Logger.out_err("")


def main():
    args = docopt(USAGE, version=VERSION)

    if args['help'] or args['--help']:
        print(USAGE)
        return True

    if args['print_supported_formats']:
        print('"native": native format')
        print('"csv"   : CSV format')
        print('          {}'.format(Noder.CSV_HEADER))
        return True

    # check format
    fmt = args['--format']
    if fmt != 'native' and fmt != 'csv':
        Logger.err('bad format: {}'.format(fmt))
        return False

    if args['--verbose']:
        print(args)

    # print banner
    if not args['--no-banner']:
        banner()

    # set colors
    if args['--no-color']:
        Logger.no_color()

    # init noder
    noder = Noder(debug=args['--verbose'], sortsize=args['--sortsize'],
                  arc=args['--archive'])
    # init catalog
    catalog = Catalog(args['--catalog'], debug=args['--verbose'],
                      force=args['--force'])
    # init top node
    top = catalog.restore()
    if not top:
        top = noder.new_top_node()

    # handle the meta node
    meta = noder.update_metanode(top)
    catalog.set_metanode(meta)

    # parse command
    if args['index']:
        cmd_index(args, noder, catalog, top)
    if args['update']:
        cmd_update(args, noder, catalog, top)
    elif args['find']:
        cmd_find(args, noder, top)
    elif args['tree']:
        cmd_tree(args, noder, top)
    elif args['ls']:
        cmd_ls(args, noder, top)
    elif args['rm']:
        cmd_rm(args, noder, catalog, top)
    elif args['graph']:
        cmd_graph(args, noder, top)
    elif args['rename']:
        cmd_rename(args, noder, catalog, top)
    elif args['edit']:
        cmd_edit(args, noder, catalog, top)

    return True


if __name__ == '__main__':
    '''entry point'''
    if main():
        sys.exit(0)
    sys.exit(1)



from setuptools import setup, find_packages
from codecs import open
from os import path
import catcli

readme = 'README.md'
here = path.abspath(path.dirname(__file__))
read_readme = lambda f: open(f, 'r').read()

VERSION = catcli.__version__
REQUIRES_PYTHON = '>=3'

setup(
    name='catcli',
    version=VERSION,

    description='The command line catalog tool for your offline data',
    long_description=read_readme(readme),
    long_description_content_type='text/markdown',
    license_files = ('LICENSE',),
    url='https://github.com/deadc0de6/catcli',
    download_url = 'https://github.com/deadc0de6/catcli/archive/v'+VERSION+'.tar.gz',
    options={"bdist_wheel": {"python_tag": "py3"}},
    # include anything from MANIFEST.in
    include_package_data=True,

    author='deadc0de6',
    author_email='deadc0de6@foo.bar',

    license='GPLv3',
    python_requires=REQUIRES_PYTHON,
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          ],

    keywords='catalog commandline indexer offline',
    packages=find_packages(exclude=['tests*']),
    install_requires=['docopt', 'anytree'],

    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage', 'pytest', 'pytest-cov'],
    },

    entry_points={
        'console_scripts': [
            'catcli=catcli:main',
        ],
    },
)
